module integrator
  real(kind=8), parameter :: IDLETIME = 0.1
  integer, parameter :: TAG_WORK = 0
  integer, parameter :: TAG_END = 2
contains

  function func(x)
    use mpi
    implicit none
    real(kind=8) :: func
    real(kind=8), intent(in) :: x

    real(kind=8), parameter :: IDLETIME = 0.1
    real(kind=8) :: t

    t = MPI_WTIME()
    ! Introduce work imballance by sleeping more at larger x
    do while (MPI_WTIME()-t <= IDLETIME*x*x)
    end do
    func = 4.0_8 / (1.0_8 + x*x)
  end function func

  function integrate(f, x_start, x_end, maxSteps) result(sum)
    use mpi
    implicit none
    interface
       function f(x)
         real(kind=8) :: f, x
       end function f
    end interface
    real(kind=8), intent(in) :: x_start, x_end
    integer, intent(in) :: maxSteps

    integer :: myRank, numProcs, ierr
    real(kind=8) :: sum, stepSize
    real(kind=8), dimension(2) :: x, y
    integer :: step, nextRank, items
    integer, dimension(MPI_STATUS_SIZE) :: status

    call MPI_COMM_RANK(MPI_COMM_WORLD, myRank, ierr)
    call MPI_COMM_SIZE(MPI_COMM_WORLD, numProcs, ierr)

    sum = 0.0
    
    stepSize = (x_end - x_start)/maxSteps

    if (myRank == 0) then
       ! I am the controller, distribute the work

       ! Part 1: Send work to everyone
       step = 0
       do while (step < numProcs-1 .and. step < maxSteps)
          x(1) = x_start + stepSize*step
          x(2) = x_start + stepSize*(step+1)
          nextRank = step+1
          call MPI_SEND(x, 2, MPI_REAL8, nextRank, TAG_WORK, &
               MPI_COMM_WORLD, ierr)
          step = step + 1
       end do

       ! Part 2: Receive 'maxSteps' result items from our workers
       items = 0
       do while (items < maxSteps)
          ! Receive a work item from anyone
          call MPI_RECV(y, 2, MPI_REAL8, MPI_ANY_SOURCE, TAG_WORK, &
               MPI_COMM_WORLD, status, ierr)
          sum = sum + stepSize*0.5_8*(y(1)+y(2))
          ! Next work item goes to the same worker - it is now fre
          nextRank = status(MPI_SOURCE)

          ! Send some work if there is still any left
          if (step < maxSteps) then
             x(1) = x_start + stepSize*step
             x(2) = x_start + stepSize*(step+1)
             call MPI_SEND(x, 2, MPI_REAL8, nextRank, TAG_WORK, &
                  MPI_COMM_WORLD, ierr)
             step = step + 1
          end if

          items = items + 1
       end do

       ! Signal workers to stop by sending empty messages with tag TAG_END
       do nextRank = 1, numProcs-1
          call MPI_SEND(nextRank, 0, MPI_INTEGER, nextRank, TAG_END, &
               MPI_COMM_WORLD, ierr)
       end do
    else
       work_loop: do while(.true.)
          ! I am worker, wait for work

          ! Receive the left and right points of the trapezoid and compute
          ! the corresponding function values. If the tag is TAG_END, don't
          ! compute but exit.
          call MPI_RECV(x, 2, MPI_REAL8, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &
               status, ierr)
          if (status(MPI_TAG) == TAG_END) exit work_loop
          y(1) = f(x(1))
          y(2) = f(x(2))
          ! Send back the computed result
          call MPI_SEND(y, 2, MPI_REAL8, 0, TAG_WORK, MPI_COMM_WORLD, ierr)
       end do work_loop
    end if
  end function integrate

end module integrator

program calc_pi
  use mpi
  use integrator
  implicit none
  integer :: myRank, numProcs, ierr
  integer :: maxSteps = 100
  real(kind=8) :: startTime, stopTime, pi
  ! Integration domain si [0, 1]
  real(kind=8), parameter :: x0 = 0.0, x1 = 1.0

  call MPI_INIT(ierr)
  call MPI_COMM_RANK(MPI_COMM_WORLD, myRank, ierr)
  call MPI_COMM_SIZE(MPI_COMM_WORLD, numProcs, ierr)

  if (myRank == 0) then
     print '("Integrating from ",F9.6," to ",F9.6," in ",I0," steps")', &
          x0, x1, maxSteps
  end if

  ! Synchronize before making performance measurements
  call MPI_BARRIER(MPI_COMM_WORLD, ierr)

  startTime = MPI_WTIME()

  pi = integrate(func, x0, x1, maxSteps)

  stopTime = MPI_WTIME()

  if (myRank == 0) then
     print '(/,"PI = ",F9.6,"    integral = ",F9.6)', 4.0_8*atan(1.0_8), pi
     print '("Computation took ",F10.6," seconds")', stopTime-startTime
  end if

  call MPI_FINALIZE(ierr)
end program calc_pi
